"""
Django Typer app config. This module includes settings check and rich traceback
installation logic.
"""

import inspect
import os
import typing as t
from types import ModuleType

from django.apps import AppConfig
from django.conf import settings
from django.core.checks import CheckMessage
from django.core.checks import Warning as CheckWarning
from django.core.checks import register
from django.utils.translation import gettext as _

from django_typer import patch
from django_typer.utils import traceback_config

patch.apply()

rich: t.Union[ModuleType, None] = None

try:
    import sys

    import rich
    from rich import traceback
    from typer import main as typer_main

    tb_config = traceback_config()
    if rich and isinstance(tb_config, dict) and not tb_config.get("no_install", False):
        # install rich tracebacks if we've been configured to do so (default)
        no_color = "NO_COLOR" in os.environ
        force_color = "FORCE_COLOR" in os.environ
        traceback.install(
            console=tb_config.pop(
                "console",
                (
                    rich.console.Console(
                        stderr=True,
                        no_color=no_color,
                        force_terminal=(
                            False if no_color else force_color if force_color else None
                        ),
                    )
                    if no_color or force_color
                    else None
                ),
            ),
            **{
                param: value
                for param, value in tb_config.items()
                if param in set(inspect.signature(traceback.install).parameters.keys())
            },
        )
        # typer installs its own exception hook and it falls back to the sys hook - depending
        # on when typer was imported it may have the original fallback system hook or our
        # installed rich one - we patch it here to make sure!
        typer_main._original_except_hook = sys.excepthook
except ImportError:
    pass


@register("settings")
def check_traceback_config(app_configs, **kwargs) -> t.List[CheckMessage]:
    """
    A system check that validates that the traceback config is valid and
    contains only the expected parameters.
    """
    warnings: t.List[CheckMessage] = []
    tb_cfg = traceback_config()
    if isinstance(tb_cfg, dict):
        if rich and traceback:
            expected = {
                "no_install",
                "short",
                *inspect.signature(traceback.install).parameters.keys(),
            }
            unexpected = set(tb_cfg.keys()) - expected
            if unexpected:
                warnings.append(
                    CheckWarning(
                        "DT_RICH_TRACEBACK_CONFIG",
                        hint=_("Unexpected parameters encountered: {keys}.").format(
                            keys=", ".join(unexpected)
                        ),
                        obj=settings.SETTINGS_MODULE,
                        id="django_typer.W001",
                    )
                )
    return warnings


class DjangoTyperConfig(AppConfig):
    """
    Django Typer app config.
    """

    name = "django_typer"
    label = name
    verbose_name = "Django Typer"



"""
A collection of useful utilities.
"""

import os
import shutil
import sys
import typing as t
from pathlib import Path
from threading import local

from django.conf import settings

# DO NOT IMPORT ANYTHING FROM TYPER HERE - SEE patch.py


def get_usage_script(script: t.Optional[str] = None) -> t.Union[Path, str]:
    """
    Return the script name if it is on the path or the absolute path to the script
    if it is not.

    :param script: The script name to check. If None the current script is used.
    :return: The script name or the relative path to the script from cwd.
    """
    cmd_pth = Path(script or sys.argv[0])
    if shutil.which(cmd_pth.name):
        return cmd_pth.name
    try:
        return cmd_pth.absolute().relative_to(Path(os.getcwd()))
    except ValueError:
        return cmd_pth.absolute()


def traceback_config() -> t.Union[bool, t.Dict[str, t.Any]]:
    """
    Fetch the rich traceback installation parameters from our settings. By default
    rich tracebacks are on with show_locals = True. If the config is set to False
    or None rich tracebacks will not be installed even if the library is present.

    This allows us to have a common traceback configuration for all commands. If rich
    tracebacks are managed separately this setting can also be switched off.
    """
    cfg = getattr(settings, "DT_RICH_TRACEBACK_CONFIG", {"show_locals": True})
    if cfg:
        return {"show_locals": True, **cfg}
    return bool(cfg)


_command_context = local()


def get_current_command() -> t.Optional["TyperCommand"]:  # type: ignore
    """
    Returns the current typer command. This can be used as a way to
    access the current command object from anywhere if we are executing
    inside of one from higher on the stack. We primarily need this because certain
    monkey patches are required in typer code - namely for enabling/disabling
    color based on configured parameters.

    This function is thread safe.

    This is analogous to click's get_current_context but for
    command execution.

    :return: The current typer command or None if there is no active command.
    """
    try:
        return t.cast("TyperCommand", _command_context.stack[-1])  # type: ignore
    except (AttributeError, IndexError):
        pass
    return None


T = t.TypeVar("T")  # pylint: disable=C0103


def with_typehint(baseclass: t.Type[T]) -> t.Type[T]:
    """
    Type hinting mixin inheritance is really annoying. The current
    canonical way is to use Protocols but this is prohibitive when
    the super classes already exist and are extensive. All we're
    trying to do is let our type checker know about super() methods
    etc - this is a simple way to do that.
    """
    if t.TYPE_CHECKING:
        return baseclass  # pragma: no cover
    return object  # type: ignore



"""
Typer_ supports custom parsers for options and arguments. If you would
like to type a parameter with a type that isn't supported by Typer_ you can
`implement your own parser <https://typer.tiangolo.com/tutorial/parameter-types/custom-types>`_
, or `ParamType <https://click.palletsprojects.com/en/8.1.x/api/#click.ParamType>`_
in click_ parlance.

This module contains a collection of parsers that turn strings into useful
Django types. Pass these parsers to the `parser` argument of typer.Option and
typer.Argument. Parsers are provided for:

- **Model Objects**: Turn a string into a model object instance using :class:`ModelObjectParser`.
- **App Labels**: Turn a string into an AppConfig instance using :func:`parse_app_label`.


.. warning::

    If you implement a custom parser, please take care to ensure that it:
        - Handles the case where the value is already the expected type.
        - Returns None if the value is None (already implemented if subclassing ParamType).
        - Raises a CommandError if the value is invalid.
        - Handles the case where the param and context are None.
"""

import typing as t
from uuid import UUID

from click import Context, Parameter, ParamType
from django.apps import AppConfig, apps
from django.core.management import CommandError
from django.db.models import Field, ForeignObjectRel, Model, UUIDField
from django.utils.translation import gettext as _

from django_typer.completers import ModelObjectCompleter


class ModelObjectParser(ParamType):
    """
    A parser that will turn strings into model object instances based on the
    configured lookup field and model class.

    .. code-block:: python

        from django_typer.parsers import ModelObjectParser

        class Command(TyperCommand):
            def handle(
                self,
                django_apps: Annotated[
                    t.List[MyModel],
                    typer.Argument(
                        parser=ModelObjectParser(MyModel, lookup_field="name"),
                        help=_("One or more application labels."),
                    ),
                ],
            ):

    .. note::

        Typer_ does not respect the shell_complete functions on ParamTypes passed as
        parsers. To add shell_completion see :class:`~django_typer.completers.ModelObjectCompleter`
        or the :func:`~django_typer.model_parser_completer` convenience
        function.

    :param model_cls: The model class to use for lookup.
    :param lookup_field: The field to use for lookup. Defaults to 'pk'.
    :param on_error: A callable that will be called if the lookup fails.
        The callable should accept three arguments: the model class, the
        value that failed to lookup, and the exception that was raised.
        If not provided, a CommandError will be raised.
    """

    error_handler = t.Callable[[t.Type[Model], str, Exception], None]

    model_cls: t.Type[Model]
    lookup_field: str
    case_insensitive: bool = False
    on_error: t.Optional[error_handler] = None

    _lookup: str = ""
    _field: Field
    _completer: ModelObjectCompleter

    __name__: str = "MODEL"  # typer internals expect this

    def __init__(
        self,
        model_cls: t.Type[Model],
        lookup_field: t.Optional[str] = None,
        case_insensitive: bool = case_insensitive,
        on_error: t.Optional[error_handler] = on_error,
    ):
        from django.contrib.contenttypes.fields import GenericForeignKey

        self.model_cls = model_cls
        self.lookup_field = str(
            lookup_field or getattr(self.model_cls._meta.pk, "name", "id")
        )
        self.on_error = on_error
        self.case_insensitive = case_insensitive
        field = self.model_cls._meta.get_field(self.lookup_field)
        assert not isinstance(field, (ForeignObjectRel, GenericForeignKey)), _(
            "{cls} is not a supported lookup field."
        ).format(cls=self._field.__class__.__name__)
        self._field = field
        if self.case_insensitive and "iexact" in self._field.get_lookups():
            self._lookup = "__iexact"
        self.__name__ = (
            str(self.model_cls._meta.verbose_name)
            if self.model_cls._meta.verbose_name
            else self.model_cls.__name__
        )

    def convert(
        self, value: t.Any, param: t.Optional[Parameter], ctx: t.Optional[Context]
    ):
        """
        Invoke the parsing action on the given string. If the value is
        already a model instance of the expected type the value will
        be returned. Otherwise the value will be treated as a value to query
        against the lookup_field. If no model object is found the error
        handler is invoked if one was provided.

        :param value: The value to parse.
        :param param: The parameter that the value is associated with.
        :param ctx: The context of the command.
        :raises CommandError: If the lookup fails and no error handler is
            provided.
        """
        try:
            if isinstance(value, self.model_cls):
                return value
            if isinstance(self._field, UUIDField):
                uuid = ""
                for char in value:
                    if char.isalnum():
                        uuid += char
                value = UUID(uuid)
            return self.model_cls.objects.get(
                **{f"{self.lookup_field}{self._lookup}": value}
            )
        except (self.model_cls.DoesNotExist, ValueError) as err:
            if self.on_error:
                return self.on_error(self.model_cls, str(value), err)
            raise CommandError(
                _('{model} "{value}" does not exist!').format(
                    model=self.model_cls.__name__, value=value
                )
            ) from err


def parse_app_label(label: t.Union[str, AppConfig]):
    """
    A parser for app labels. If the label is already an AppConfig instance,
    the instance is returned. The label will be tried first, if that fails
    the label will be treated as the app name.

    .. code-block:: python

        import typing as t
        import typer
        from django_typer import TyperCommand
        from django_typer.parsers import parse_app_label

        class Command(TyperCommand):

            def handle(
                self,
                django_apps: t.Annotated[
                    t.List[AppConfig],
                    typer.Argument(
                        parser=parse_app_label,
                        help=_("One or more application labels.")
                    )
                ]
            ):
                ...

    :param label: The label to map to an AppConfig instance.
    :raises CommandError: If no matching app can be found.
    """
    if isinstance(label, AppConfig):
        return label
    try:
        return apps.get_app_config(label)
    except LookupError as err:
        for cfg in apps.get_app_configs():
            if cfg.name == label:
                return cfg

        raise CommandError(
            _("{label} does not match any installed app label.").format(label=label)
        ) from err


parse_app_label.__name__ = "APP_LABEL"



"""
The shellcompletion command is a Django_ management command that installs and removes
shellcompletion scripts for supported shells (bash, fish, zsh, powershell). This
command is also the entry point for running the completion logic and can be used to
debug completer code.

.. typer:: django_typer.management.commands.shellcompletion.Command:typer_app
    :prog: manage.py shellcompletion
    :width: 80
    :convert-png: latex

:func:`~django_typer.management.commands.shellcompletion.Command.install` invokes typer's shell
completion installation logic, but does have to patch the installed scripts. This is because
there is only one installation for all Django_ management commands, not each individual command.
The completion logic here will failover to Django_'s builtin autocomplete if the command in
question is not a :class:`~django_typer.TyperCommand`. To promote compatibility with other
management command libraries or custom completion logic, a fallback completion function can also
be specified.
"""

# A needed refactoring here would be to provide root hooks for completion logic in Django core that
# base classes can register for. This would provide a coordinated way for libraries like
# django-typer to plug in their own completion logic.

import contextlib
import inspect
import io
import os
import sys
import typing as t
from pathlib import Path

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from functools import cached_property

from click.parser import split_arg_string
from click.shell_completion import (
    CompletionItem,
    add_completion_class,
    get_completion_class,
)
from django.core.management import CommandError, ManagementUtility
from django.utils.module_loading import import_string
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from shellingham import ShellDetectionFailure, detect_shell
from typer import Argument, Option, echo
from typer.completion import Shells  # type: ignore
from typer.completion import completion_init

from django_typer import TyperCommand, command, get_command
from django_typer.utils import get_usage_script

DETECTED_SHELL = None

try:
    DETECTED_SHELL = detect_shell()[0]
except (ShellDetectionFailure, RuntimeError):
    pass

DJANGO_COMMAND = Path(__file__).name.split(".")[0]


class Command(TyperCommand):
    """
    This command installs autocompletion for the current shell. This command uses the typer/click
    autocompletion scripts to generate the autocompletion items, but monkey patches the scripts
    to invoke our bundled shell complete script which fails over to the django autocomplete
    function when the command being completed is not a :class:`~django_typer.TyperCommand`. When
    the django autocomplete function is used we also wrap it so that it works for any supported
    click/typer shell, not just bash.

    We also provide a remove command to easily remove the installed script.

    Great pains are taken to use the upstream dependency's shell completion logic. This is so
    advances and additional shell support implemented upstream should just work. However, it
    would be possible to add support for new shells here using the pluggable logic that click
    provides. It is probably a better idea however to add support for new shells at the typer
    level.

    Shell autocompletion can be brittle with every shell having its own quirks and nuances. We
    make a good faith effort here to support all the shells that typer/click support, but there
    can easily be system specific configuration issues that prevent this from working. In those
    cases users should refer to the online documentation for their specific shell to troubleshoot.
    """

    help = t.cast(str, _("Install autocompletion for the current shell."))

    # disable the system checks - no reason to run these for this one-off command
    requires_system_checks = []
    requires_migrations_checks = False

    # remove unnecessary django command base parameters - these just clutter the help
    suppressed_base_arguments = {
        "version",
        "skip_checks",
        "no_color",
        "force_color",
        "verbosity",
    }

    _shell: Shells

    COMPLETE_VAR = "_COMPLETE_INSTRUCTION"

    @cached_property
    def manage_script(self) -> t.Union[str, Path]:
        """
        Returns the name of the manage command as a string if it is available as a command
        on the user path. If it is a script that is not available as a command on the path
        it will return an absolute Path object to the script.
        """
        # We do not want to hardcode this as 'manage.py' because users are free to rename and
        # implement their own manage scripts! The safest way to do this is therefore to fetch
        # the name of the script that invoked the shellcompletion command and determine if that
        # script is available on the user's path or not.

        # Most shell's completion infrastructure works best if the commands are available on the
        # path. However, it is common for Django development to be done in a virtual environment
        # with a manage.py script being invoked directly as a script. Completion should work in
        # this case as well, but it does complicate the installation for some shell's so we must
        # first figure out which mode we are in.
        script = get_usage_script()
        if isinstance(script, Path):
            return script.absolute()
        return script

    @cached_property
    def manage_script_name(self) -> str:
        """
        Get the name of the manage script as a command available from the shell's path.
        """
        return str(getattr(self.manage_script, "name", self.manage_script))

    @property
    def shell(self) -> Shells:
        """
        Get the active shell. If not explicitly set, it first tries to find the shell
        in the environment variable shell complete scripts set and failing that it will try
        to autodetect the shell.
        """
        return getattr(
            self,
            "_shell",
            (
                Shells(os.environ[self.COMPLETE_VAR].partition("_")[2])
                if self.COMPLETE_VAR in os.environ
                else None
            ),
        ) or Shells(detect_shell()[0])

    @shell.setter
    def shell(self, shell: t.Optional[Shells]):
        """Set the shell to install autocompletion for."""
        if shell is None:
            try:
                self._shell = Shells(detect_shell()[0])
            except (ShellDetectionFailure, RuntimeError) as err:
                raise CommandError(
                    gettext(
                        "Please specify the shell to install or remove "
                        "autocompletion for. Unable to detect shell: {err}"
                    ).format(err=str(err))
                ) from err
        else:
            self._shell = shell if isinstance(shell, Shells) else Shells(shell)

    @cached_property
    def noop_command(self):
        """
        This is a no-op command that is used to bootstrap click Completion classes. It
        has no use other than to avoid any potential attribute errors when we emulate
        upstream completion logic
        """
        return self.get_subcommand("noop")

    def patch_script(
        self, shell: t.Optional[Shells] = None, fallback: t.Optional[str] = None
    ) -> None:
        """
        We have to monkey patch the typer completion scripts to point to our custom
        shell complete script. This is potentially brittle but thats why we have robust
        CI!

        :param shell: The shell to patch the completion script for.
        :param fallback: The python import path to a fallback autocomplete function to use when
            the completion command is not a TyperCommand. Defaults to None, which means the bundled
            complete script will fallback to the django autocomplete function, but wrap it so it
            works for all supported shells other than just bash.
        """
        # do not import this private stuff until we need it - avoids it tanking the whole
        # library if these imports change
        from typer import (
            _completion_shared as typer_scripts,  # pylint: disable=import-outside-toplevel
        )

        shell = shell or self.shell

        fallback = f" --fallback {fallback}" if fallback else ""

        def replace(s: str, old: str, new: str, occurrences: t.List[int]) -> str:
            """
            :param s: The string to modify
            :param old: The string to replace
            :param new: The string to replace with
            :param occurrences: A list of occurrences of the old string to replace with the
                new string, where the occurrence number is the zero-based count of the old
                strings appearance in the string when counting from the front.
            """
            count = 0
            result = ""
            start = 0

            for end in range(len(s)):
                if s[start : end + 1].endswith(old):
                    if count in occurrences:
                        result += f"{s[start:end+1-len(old)]}{new}"
                        start = end + 1
                    else:
                        result += s[start : end + 1]
                        start = end + 1
                    count += 1

            result += s[start:]
            return result

        if shell is Shells.bash:
            typer_scripts._completion_scripts[Shells.bash.value] = replace(
                typer_scripts.COMPLETION_SCRIPT_BASH,
                "$1",
                f"$1 {DJANGO_COMMAND} complete",
                [0],
            )
        elif shell is Shells.zsh:
            typer_scripts._completion_scripts[Shells.zsh.value] = replace(
                typer_scripts.COMPLETION_SCRIPT_ZSH,
                "%(prog_name)s",
                f"${{words[0,1]}} {DJANGO_COMMAND} complete",
                [1],
            )
        elif shell is Shells.fish:
            typer_scripts._completion_scripts[Shells.fish.value] = replace(
                typer_scripts.COMPLETION_SCRIPT_FISH,
                "%(prog_name)s",
                f"{self.manage_script} {DJANGO_COMMAND} complete",
                [1, 2],
            )
        else:
            assert shell in [
                Shells.pwsh,
                Shells.powershell,
            ], gettext(
                "Unsupported shell: {shell}"
            ).format(shell=shell.value)
            script = replace(
                typer_scripts.COMPLETION_SCRIPT_POWER_SHELL,
                "%(prog_name)s",
                f"{self.manage_script} {DJANGO_COMMAND} complete",
                [0],
            )
            typer_scripts._completion_scripts[Shells.powershell.value] = script
            typer_scripts._completion_scripts[Shells.pwsh.value] = script

    @command(
        help=t.cast(str, _("Install autocompletion for the current or given shell."))
    )
    def install(
        self,
        shell: Annotated[
            t.Optional[Shells],
            Argument(
                help=t.cast(
                    str, _("Specify the shell to install or remove autocompletion for.")
                )
            ),
        ] = DETECTED_SHELL,
        manage_script: Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The name of the django manage script to install autocompletion for if "
                        "different than the script used to invoke this command."
                    ),
                )
            ),
        ] = None,
        fallback: Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The python import path to a fallback complete function to use when "
                        "the completion command is not a TyperCommand."
                    ),
                )
            ),
        ] = None,
    ):
        """
        Install autocompletion for the given shell. If the shell is not specified, it will
        try to detect the shell. If the shell is not detected, it will fail.

        We run the upstream typer installation routines, with some augmentation.

        .. typer:: django_typer.management.commands.shellcompletion.Command:typer_app:install
            :width: 85
            :convert-png: latex
        """
        # do not import this private stuff until we need it - avoids tanking the whole
        # library if these imports change
        from typer._completion_shared import install

        self.shell = shell  # type: ignore
        self.patch_script(fallback=fallback)
        install_path = install(
            shell=self.shell.value,
            prog_name=str(manage_script or self.manage_script_name),
            complete_var=self.COMPLETE_VAR,
        )[1]
        self.stdout.write(
            self.style.SUCCESS(  # pylint: disable=no-member
                gettext("Installed autocompletion for {shell} @ {install_path}").format(
                    shell=self.shell.value, install_path=install_path
                )
            )
        )

    @command(
        help=t.cast(str, _("Remove autocompletion for the current or given shell."))
    )
    def remove(
        self,
        shell: Annotated[
            t.Optional[Shells],
            Argument(
                help=t.cast(
                    str,
                    _("Specify the shell to install or remove shell completion for."),
                )
            ),
        ] = DETECTED_SHELL,
        manage_script: Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The name of the django manage script to remove shell completion for if "
                        "different than the script used to invoke this command."
                    ),
                )
            ),
        ] = None,
    ):
        """
        Remove the autocompletion for the given shell. If the shell is not specified, it will
        try to detect the shell. If the shell is not detected, it will fail.

        Since the installation routine is upstream we first run install to determine where the
        completion script is installed and then we remove it.

        .. typer:: django_typer.management.commands.shellcompletion.Command:typer_app:remove
            :prog: shellcompletion
            :width: 80
            :convert-png: latex

        """
        # do not import this private stuff until we need it - avoids tanking the whole
        # library if these imports change
        from typer._completion_shared import install

        # its less brittle to install and use the returned path to uninstall
        self.shell = shell  # type: ignore
        prog_name = str(manage_script or self.manage_script_name)
        installed_path = install(
            shell=self.shell.value if self.shell else None, prog_name=prog_name
        )[1]
        if self.shell in [Shells.pwsh, Shells.powershell]:
            # annoyingly, powershell has one profile script for all completion commands
            # so we have to find our entry and remove it
            edited_lines = []
            mark = None
            with open(installed_path, "rt", encoding="utf-8") as pwr_sh:
                for line in pwr_sh.readlines():
                    edited_lines.append(line)
                    if line.startswith("Import-Module PSReadLine"):
                        mark = len(edited_lines) - 1
                    elif (
                        mark is not None
                        and line.startswith("Register-ArgumentCompleter")
                        and f" {prog_name} " in line
                    ):
                        edited_lines = edited_lines[:mark]
                        mark = None

            if edited_lines:
                with open(installed_path, "wt", encoding="utf-8") as pwr_sh:
                    pwr_sh.writelines(edited_lines)
            else:
                installed_path.unlink()

        else:
            installed_path.unlink()
            rc_file = {
                None: None,
                Shells.bash: Path("~/.bashrc").expanduser(),
                Shells.zsh: Path("~/.zshrc").expanduser(),
            }.get(self.shell, None)
            if rc_file and rc_file.is_file():
                edited: t.List[str] = []
                with open(rc_file, "rt", encoding="utf-8") as rc:
                    for line in rc.readlines():
                        if (
                            self.shell is Shells.bash
                            and line.strip() == f"source {installed_path}"
                        ):
                            continue
                        edited.append(line)
                # remove empty lines from the end of the file, the typer install scripts add
                # extra newlines
                while edited and not edited[-1].strip():
                    edited.pop()
                edited.append("")  # add one back on
                with open(rc_file, "wt", encoding="utf-8") as rc:
                    rc.writelines(edited)

        self.stdout.write(
            self.style.WARNING(  # pylint: disable=no-member
                gettext("Removed autocompletion for {shell}.").format(
                    shell=self.shell.value if self.shell else ""
                )
            )
        )

    @command(
        help=t.cast(str, _("Generate autocompletion for command string.")), hidden=False
    )
    def complete(
        self,
        cmd_str: Annotated[
            t.Optional[str],
            Argument(
                metavar="command",
                help=t.cast(
                    str, _("The command string to generate completion suggestions for.")
                ),
            ),
        ] = None,
        shell: Annotated[
            t.Optional[Shells],
            Option(
                help=t.cast(
                    str,
                    _(
                        "Specify the shell to fetch completion for, default will autodetect."
                    ),
                )
            ),
        ] = None,
        fallback: Annotated[
            t.Optional[str],
            Option(
                help=t.cast(
                    str,
                    _(
                        "The python import path to a fallback complete function to use when "
                        "the completion command is not a TyperCommand. By default, the builtin "
                        "django autocomplete function is used."
                    ),
                )
            ),
        ] = None,
    ):
        """
        We implement the shell complete generation script as a Django command because the
        Django environment needs to be bootstrapped for it to work. This also allows
        us to test completion logic in a platform agnostic way.

        .. tip::

            This command is super useful for debugging shell_complete logic. For example to
            enter into the debugger, we could set a breakpoint in our ``shell_complete`` function
            for the option parameter and then run the following command:

            .. code-block:: bash

                $ ./manage.py shellcompletion complete "./manage.py your_command --option "

        .. typer:: django_typer.management.commands.shellcompletion.Command:typer_app:complete
            :prog: shellcompletion
            :width: 80
            :convert-png: latex
        """
        os.environ[self.COMPLETE_VAR] = (
            f"complete_{shell.value}"
            if shell
            else os.environ.get(self.COMPLETE_VAR, f"complete_{self.shell.value}")
        )
        self.shell = Shells(os.environ[self.COMPLETE_VAR].partition("_")[2])

        completion_init()
        CompletionClass = get_completion_class(  # pylint: disable=C0103
            self.shell.value
        )
        assert CompletionClass, gettext(
            'No completion implementation for "{shell}"'
        ).format(shell=self.shell.value)
        if cmd_str:
            # when the command is given, this is a user testing their autocompletion,
            # so we need to override the completion classes get_completion_args logic
            # because our entry point was not an installed completion script
            def get_completion_args(self) -> t.Tuple[t.List[str], str]:
                cwords = split_arg_string(cmd_str)
                if cmd_str[-1].isspace():
                    cwords.append("")
                # allow users to not specify the manage script, but allow for it
                # if they do by lopping it off - same behavior as upstream classes
                try:
                    if Path(cwords[0]).resolve() == Path(sys.argv[0]).resolve():
                        cwords = cwords[1:]
                except (TypeError, ValueError, OSError):  # pragma: no cover
                    pass
                return (
                    cwords[:-1],
                    cwords[-1] if cwords else "",
                )

            CompletionClass.get_completion_args = get_completion_args  # type: ignore

        # guard against double patching
        if not getattr(CompletionClass, "_dt_patched", False):
            _get_completions = CompletionClass.get_completions

            def get_completions(self, args, incomplete):
                """
                need to remove the django command name from the arg completions
                """
                return _get_completions(self, args[1:], incomplete)

            CompletionClass.get_completions = get_completions  # type: ignore

            add_completion_class(CompletionClass, self.shell.value)
            CompletionClass._dt_patched = True  # type: ignore

        args = CompletionClass(
            cli=self.noop_command.click_command,
            ctx_args={},
            prog_name=sys.argv[0],
            complete_var=self.COMPLETE_VAR,
        ).get_completion_args()[0]

        def call_fallback(fb: t.Optional[str]) -> None:
            fallback = import_string(fb) if fb else self.django_fallback
            if cmd_str and inspect.signature(fallback).parameters:
                fallback(cmd_str)  # pyright: ignore[reportCallIssue]
            else:
                fallback()

        def get_completion() -> None:
            if not args:
                call_fallback(fallback)
            else:
                # we first need to figure out which command is being invoked
                # but we cant be sure which index the command name is at so
                # we try to fetch each in order and assume the first arg
                # that resolves to a command is the command
                cmd = None
                cmd_idx = -1
                try:
                    while cmd is None:
                        cmd_idx += 1
                        try:
                            cmd = get_command(args[cmd_idx])
                        except KeyError:
                            pass
                except IndexError as err:
                    if cmd_str:
                        # if we're here a user is probably trying to debug a completion
                        # notify them that the command is not recognized
                        raise CommandError(
                            gettext('Unrecognized command: "{cmd_str}"').format(
                                cmd_str=cmd_str
                            )
                        ) from err
                    raise  # otherwise nowhere to go - just error out

                if isinstance(cmd, TyperCommand):  # type: ignore[unreachable]
                    # this will exit out so no return is needed here
                    cmd.typer_app(
                        args=args[cmd_idx + 1 :],
                        standalone_mode=True,
                        django_command=cmd,
                        complete_var=self.COMPLETE_VAR,
                        prog_name=(
                            f"{' '.join(sys.argv[0:cmd_idx or None])} "
                            f"{self.typer_app.info.name}",
                        ),
                    )
                call_fallback(fallback)

        try:
            get_completion()
        except SystemExit:
            if cmd_str:
                return
            raise

    def django_fallback(self):
        """
        Run django's builtin bash autocomplete function. We wrap the click
        completion class to make it work for all supported shells, not just
        bash.
        """
        CompletionClass = get_completion_class(  # pylint: disable=C0103
            self.shell.value
        )
        assert CompletionClass, gettext(
            'No completion implementation for "{shell}"'
        ).format(shell=self.shell.value)

        def get_completions(self, args, incomplete):
            # spoof bash environment variables
            # the first one is lopped off, so we insert a placeholder 0
            args = ["0", *args]
            if args[-1] != incomplete:
                args.append(incomplete)
            else:  # pragma: no cover
                pass
            os.environ["COMP_WORDS"] = " ".join(args)
            os.environ["COMP_CWORD"] = str(args.index(incomplete))
            os.environ["DJANGO_AUTO_COMPLETE"] = "1"
            dj_manager = ManagementUtility(args)
            capture_completions = io.StringIO()
            with contextlib.redirect_stdout(capture_completions):
                try:
                    dj_manager.autocomplete()
                except SystemExit:
                    pass
            return [
                CompletionItem(item)
                for item in capture_completions.getvalue().split()
                if item
            ]

        CompletionClass.get_completions = get_completions  # type: ignore
        echo(
            CompletionClass(
                cli=self.noop_command.click_command,
                ctx_args={},
                prog_name=self.manage_script_name,
                complete_var=self.COMPLETE_VAR,
            ).complete()
        )

    @command(
        hidden=True,
        context_settings={
            "ignore_unknown_options": True,
            "allow_extra_args": True,
            "allow_interspersed_args": True,
        },
    )
    def noop(self):
        """
        This is a no-op command that is used to bootstrap click Completion classes. It
        has no use other than to avoid any potential attribute access errors when we spoof
        completion logic
        """



"""
Common types for command line argument specification.
"""

# pylint: disable=pointless-string-statement, line-too-long

import sys
from pathlib import Path
from typing import Optional, cast

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from django.core.management import CommandError
from django.utils.translation import gettext_lazy as _
from typer import Option

COMMON_PANEL = "Django"


def print_version(context, _, value):
    """
    A callback to run the get_version() routine of the
    command when --version is specified.
    """
    if value:
        context.django_command.stdout.write(context.django_command.get_version())
        sys.exit()


def set_no_color(context, param, value):
    """
    If the value was provided set it on the command.
    """
    if value:
        context.django_command.no_color = value
        if context.params.get("force_color", False):
            raise CommandError(
                _("The --no-color and --force-color options can't be used together.")
            )
    return value


def set_force_color(context, param, value):
    """
    If the value was provided set it on the command.
    """
    if value:
        context.django_command.force_color = value
    return value


Version = Annotated[
    bool,
    Option(
        "--version",
        help=cast(str, _("Show program's version number and exit.")),
        callback=print_version,
        is_eager=True,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --version option <https://docs.djangoproject.com/en/stable/howto/custom-management-commands/#django.core.management.BaseCommand.get_version>`_.

The --version option is included by default and behaves the same as on BaseCommand_.
"""


Verbosity = Annotated[
    int,
    Option(
        help=cast(
            str,
            _(
                "Verbosity level; 0=minimal output, 1=normal output, "
                "2=verbose output, 3=very verbose output"
            ),
        ),
        show_choices=True,
        min=0,
        max=3,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --verbosity option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-verbosity>`_.
:class:`~django_typer.TyperCommand` does not include the verbosity option by default, but it can be
added to the command like so if needed.

.. code-block:: python
    
        from django_typer.types import Verbosity
    
        def handle(self, verbosity: Verbosity = 1):
            ...
"""

Settings = Annotated[
    str,
    Option(
        help=cast(
            str,
            _(
                "The Python path to a settings module, e.g. "
                '"myproject.settings.main". If this isn\'t provided, the '
                "DJANGO_SETTINGS_MODULE environment variable will be used."
            ),
        ),
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --settings option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-settings>`_.

The --settings option is included by default and behaves the same as on BaseCommand_ use it to
specify or override the settings module to use.
"""


PythonPath = Annotated[
    Optional[Path],
    Option(
        help=cast(
            str,
            _(
                "A directory to add to the Python path, e.g. "
                '"/home/djangoprojects/myproject".'
            ),
        ),
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --pythonpath option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-pythonpath>`_.

The --pythonpath option is included by default and behaves the same as on BaseCommand_ use it to
specify a directory to add to the Python sys path.
"""


Traceback = Annotated[
    bool,
    Option(
        "--traceback",
        help=cast(str, _("Raise on CommandError exceptions")),
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --traceback option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-traceback>`_.

The --traceback option is included by default and behaves the same as on BaseCommand_ use it to
allow CommandError exceptions to propagate out of the command and produce a stack trace.
"""


NoColor = Annotated[
    bool,
    Option(
        "--no-color",
        help=cast(str, _("Don't colorize the command output.")),
        is_eager=True,
        callback=set_no_color,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --no-color option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-no-color>`_.

The --no-color option is included by default and behaves the same as on BaseCommand_ use it to
force disable colorization of the command. You can check the supplied value of --no-color by
checking the no_color attribute of the command instance.
"""

ForceColor = Annotated[
    bool,
    Option(
        "--force-color",
        help=cast(str, _("Force colorization of the command output.")),
        is_eager=True,
        callback=set_force_color,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --force-color option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-force-color>`_.

The --force-color option is included by default and behaves the same as on BaseCommand_ use it to
force colorization of the command. You can check the supplied value of --force-color by checking
the force_color attribute of the command instance.
"""

SkipChecks = Annotated[
    bool,
    Option(
        "--skip-checks",
        help=cast(str, _("Skip system checks.")),
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the 
`Django --skip-checks option <https://docs.djangoproject.com/en/stable/ref/django-admin/#cmdoption-skip-checks>`_.

The --skip-checks option is included by default and behaves the same as on BaseCommand_ use it to
skip system checks.
"""



"""
Stuffing the typer interface into the Django BaseCommand_ interface is an exercise
in fitting a square peg into a round hole. A small amount of upstream patching is
required to make this work. All monkey patching is defined in this module to more
easily keep track of it.

Mostly because typer does not allow direct access to the rich console objects it
creates we need to do some monkey patching to get the desired behavior when
--no-color options or --force-color options are specified

apply() needs to be called before any imports from Typer

.. todo::
    
    Revisit this if Typer exposes control of the console objects.

To achieve expected --force-color and --no-color behavior there are 3 different
console objects installed when rich is used:

- The first is the global exception handler installed by rich - see apps.py
- The second is the command exception handler console installed by typer - see below
- The third is the console used by the command itself to output helps - see below

"""

import os
import sys
import typing as t

import click

from django_typer.utils import get_current_command

PATCH_APPLIED = False

# DO NOT IMPORT ANYTHING FROM TYPER HERE


def apply() -> None:
    """
    Apply monkey patches to get our console objects to recognize django's --no-color
    and --force-color options.
    """
    global PATCH_APPLIED
    if PATCH_APPLIED:
        return
    PATCH_APPLIED = True

    try:
        # Django as of <=5.0 calls colorama.init() if colorama is installed
        # this screws up forced terminals on platforms other than windows that
        # are not attached to ttys. Upstream Django should change the init
        # call to a just_fix_windows_console - we undo this and redo the right
        # thing here.
        import colorama

        colorama.deinit()
        colorama.just_fix_windows_console()
    except ImportError:
        pass

    try:
        from rich.console import Console

        # this has to go here before rich Consoles are instantiated by Typer
        color_system: t.Optional[
            t.Literal["auto", "standard", "256", "truecolor", "windows"]
        ] = "auto"
        force_terminal: t.Optional[bool] = None
        if "--no-color" in sys.argv and "--force-color" not in sys.argv:
            os.environ["NO_COLOR"] = "1"
            color_system = None
        elif "--force-color" in sys.argv:
            os.environ["FORCE_COLOR"] = "1"
            force_terminal = True

        from typer import main

        main.console_stderr = Console(
            stderr=True, force_terminal=force_terminal, color_system=color_system
        )

        # This monkey patch is required because typer does
        # not expose a good way to custom configure the Console objects
        # it uses - revisit this if/when typer exposes control of the
        # console object.
        from typer import rich_utils

        console_getter = rich_utils._get_rich_console

        def get_console(stderr: bool = False) -> Console:
            """
            Tweak the internals of the Console created by typer to match
            our context or command.

            Of all the patching this is the sketchiest.
            """
            console = console_getter(stderr=stderr)
            ctx = click.get_current_context(silent=True)
            cmd = get_current_command()
            console.no_color = (
                ctx.params.get("no_color", "NO_COLOR" in os.environ)
                if ctx
                else (cmd.no_color if cmd else "NO_COLOR" in os.environ)
            )
            console._file = (
                (cmd.stderr if stderr else cmd.stdout) if cmd else console._file
            )
            if console.no_color:
                # also remove highlights so there are
                # no ansi control characters in the text
                console._color_system = None
            elif cmd and cmd.force_color:
                console._force_terminal = True  # set this before detect color!
                console._color_system = console._detect_color_system()
            return console

        rich_utils._get_rich_console = get_console
    except ImportError:
        pass

    # this is a patch to fix lazy translation failure in some circumstances
    # when Argument helps are gettext_lazy proxies. This is I think actually
    # a problem with typer that can be fixed in a number of different ways
    # but this is the easiest
    from click import _compat

    strip_ansi = _compat.strip_ansi
    _compat.strip_ansi = lambda value: strip_ansi(str(value))


from typer import __version__ as typer_version

if (0, 4, 0) <= tuple(int(v) for v in typer_version.split(".")) <= (0, 9, 0):
    from typer import main as typer_main
    from typer.models import ParamMeta

    upstream_get_click_param = typer_main.get_click_param

    def patched_get_click_param(
        param: ParamMeta,
    ) -> t.Tuple[t.Union[click.Argument, click.Option], t.Any]:
        """
        Patch this bug: https://github.com/tiangolo/typer/issues/334

        This gets shell completions working for arguments. As of 0.9.0 I have
        an open PR to fix this problem upstream, but it is enough of an issue
        that its worth patching here until it gets merged.
        """
        click_param = upstream_get_click_param(param)
        if isinstance(click_param[0], click.Argument) and getattr(
            param.default, "shell_complete", None
        ):
            click_param[0]._custom_shell_complete = param.default.shell_complete
        return click_param

    typer_main.get_click_param = patched_get_click_param
else:  # pragma: no cover
    pass
